#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

'''


from time import time
import numpy as np
import neuron
from pathlib import Path


def _run_simulation(cell, cvode, variable_dt=False, atol=0.001):
    '''
    Running the actual simulation in NEURON, simulations in NEURON
    are now interruptable.
    '''
    neuron.h.dt = cell.dt

    # variable dt method
    if variable_dt:
        cvode.active(1)
        cvode.atol(atol)
    else:
        cvode.active(0)

    # re-initialize state
    neuron.h.finitialize(cell.v_init)

    # initialize current- and record
    if cvode.active():
        cvode.re_init()
    else:
        neuron.h.fcurrent()
    neuron.h.frecord_init()

    # Starting simulation at tstart
    neuron.h.t = cell.tstart

    cell._loadspikes()

    #print sim.time and realtime factor at intervals
    counter = 0.
    t0 = time()
    ti = neuron.h.t
    if cell.tstop >= 10000:
        interval = 1000. / cell.dt
    else:
        interval = 100. / cell.dt

    while neuron.h.t < cell.tstop:
        neuron.h.fadvance()
        counter += 1.
        if counter % interval == 0:
            rtfactor = (neuron.h.t - ti) * 1E-3 / (time() - t0)
            if cell.verbose:
                print('t = {:.0f}, realtime factor: {:.3f}'.format(neuron.h.t,
                                                                   rtfactor))
            t0 = time()
            ti = neuron.h.t


def _run_simulation_with_electrode(cell, cvode, electrode=None,
                                   variable_dt=False, atol=0.001,
                                   to_memory=True, to_file=False,
                                   file_name=None, dotprodcoeffs=None,
                                   rec_current_dipole_moment=False):
    '''
    Running the actual simulation in NEURON.
    electrode argument used to determine coefficient
    matrix, and calculate the LFP on every time step.
    '''
    try:
        import h5py
    except:
        print('h5py not found, LFP to file not possible')
        to_file = False
        file_name = None

    if variable_dt and to_file:
        print('LFP to file with variable_dt is not supported')
        to_file = False
        file_name = None

    # Use electrode object(s) to calculate coefficient matrices for LFP
    # calculations. If electrode is a list, then
    if cell.verbose:
        print('precalculating geometry - LFP mapping')

    #put electrodecoeff in a list, if it isn't already
    if dotprodcoeffs is not None:
        if type(dotprodcoeffs) != list:
            dotprodcoeffs = [dotprodcoeffs]
        electrodes = []
    else:
        #create empty list if no dotprodcoeffs are supplied
        dotprodcoeffs = []

    #just for safekeeping
    lendotprodcoeffs0 = len(dotprodcoeffs)

    #access electrode object and append mapping
    if electrode is not None:
        #put electrode argument in list if needed
        if type(electrode) == list:
            electrodes = electrode
        else:
            electrodes = [electrode]

        for el in electrodes:
            el.calc_mapping(cell)
            dotprodcoeffs.append(el.mapping)
    else:
        electrodes = None


    # Initialize NEURON simulations of cell object
    neuron.h.dt = cell.dt

    # needed for variable dt method
    if variable_dt:
        cvode.active(1)
        cvode.atol(atol)
    else:
        cvode.active(0)

    # re-initialize state
    neuron.h.finitialize(cell.v_init)

    # initialize current- and record
    if cvode.active():
        cvode.re_init()
    else:
        neuron.h.fcurrent()

    neuron.h.frecord_init()  # wrong voltages t=0 for tstart < 0 otherwise

    #Starting simulation at tstart (which may be < 0)
    neuron.h.t = cell.tstart

    #load spike times from NetCon
    cell._loadspikes()

    #print sim.time at intervals
    counter = 0.
    tstep = 0
    t0 = time()
    ti = neuron.h.t
    if cell.tstop >= 10000:
        interval = 1000. / cell.dt
    else:
        interval = 100. / cell.dt

    #temp vector to store membrane currents at each timestep
    imem = np.zeros(cell.totnsegs)
    #LFPs for each electrode will be put here during simulation

    # this will fail with cvode!
    if to_memory:
        electrodesLFP = []
        for coeffs in dotprodcoeffs:
            if not cvode.active():
                # if cvode is on lists are created on-the-fly
                electrodesLFP.append(np.zeros((coeffs.shape[0],
                                     int(cell.tstop / cell.dt) + 1)))

    #LFPs for each electrode will be put here during simulations
    if to_file:
        #ensure right ending:
        file_name = Path(file_name)
        if file_name.suffix != '.h5':
            file_name = file_name.parent / (file_name.name + '.h5')
        if not cvode.active():
            el_LFP_file = h5py.File(file_name, 'w')
            i = 0
            for coeffs in dotprodcoeffs:
                el_LFP_file['electrode{:03d}'.format(i)] = np.zeros((coeffs.shape[0], int(cell.tstop / cell.dt + 1)))
                i += 1

    # create a 2D array representation of segment midpoints for dot product
    # with transmembrane currents when computing dipole moment
    if rec_current_dipole_moment:
        midpoints = np.c_[cell.xmid, cell.ymid, cell.zmid]

    timesteps = []
    #run fadvance until time limit, and calculate LFPs for each timestep
    while neuron.h.t < cell.tstop:
        if neuron.h.t >= 0:
            i = 0
            for sec in cell.allseclist:
                for seg in sec:
                    imem[i] = seg.i_membrane_
                    i += 1

            if rec_current_dipole_moment:
                if not cvode.active():
                    cell.current_dipole_moment[tstep, ] = np.dot(imem, midpoints)
                else:
                    cell.current_dipole_moment.append(np.dot(imem, midpoints))

            if to_memory:
                for j, coeffs in enumerate(dotprodcoeffs):
                    if not cvode.active():
                        electrodesLFP[j][:, tstep] = np.dot(coeffs, imem)
                    else:
                        if tstep == 0:
                            electrodesLFP.append(np.dot(coeffs, imem)[:, np.newaxis])
                        else:
                            electrodesLFP[j] = np.hstack((electrodesLFP[j], np.dot(coeffs, imem)[:, np.newaxis]))

            if to_file:
                for j, coeffs in enumerate(dotprodcoeffs):
                    if not cvode.active():
                        el_LFP_file['electrode{:03d}'.format(j)][:, tstep] = np.dot(coeffs, imem)
            tstep += 1
        neuron.h.fadvance()
        counter += 1.
        if counter % interval == 0.:
            rtfactor = (neuron.h.t - ti) * 1E-3 / (time() - t0)
            if cell.verbose:
                print('t = {:.0f}, realtime factor: {:.3f}'.format(neuron.h.t,
                                                                   rtfactor))
            t0 = time()
            ti = neuron.h.t

    # calculate LFP after final fadvance() if needed
    if tstep < len(cell._neuron_tvec):
        i = 0
        for sec in cell.allseclist:
            for seg in sec:
                imem[i] = seg.i_membrane_
                i += 1

        if rec_current_dipole_moment:
            if not cvode.active():
                cell.current_dipole_moment[tstep, ] = np.dot(imem, midpoints)
            else:
                cell.current_dipole_moment.append(np.dot(imem, midpoints))

        if to_memory:
            for j, coeffs in enumerate(dotprodcoeffs):
                if not cvode.active():
                    electrodesLFP[j][:, tstep] = np.dot(coeffs, imem)
                else:
                    electrodesLFP[j] = np.hstack((electrodesLFP[j], np.dot(coeffs, imem)[:, np.newaxis]))

        if to_file:
            for j, coeffs in enumerate(dotprodcoeffs):
                el_LFP_file['electrode{:03d}'.format(j)][:, tstep] = np.dot(coeffs, imem)

    if rec_current_dipole_moment:
        cell.current_dipole_moment = np.array(cell.current_dipole_moment)

    # Final step, put LFPs in the electrode object, superimpose if necessary
    # If electrode.perCellLFP, store individual LFPs
    if to_memory:
        #the first few belong to input dotprodcoeffs
        cell.dotprodresults = electrodesLFP[:lendotprodcoeffs0]
        #the remaining belong to input electrode arguments
        if electrodes is not None:
            for j, LFP in enumerate(electrodesLFP):
                if not j < lendotprodcoeffs0:
                    if hasattr(electrodes[j-lendotprodcoeffs0], 'LFP'):
                        electrodes[j-lendotprodcoeffs0].LFP += LFP
                    else:
                        electrodes[j-lendotprodcoeffs0].LFP = LFP
                    #will save each cell contribution separately
                    if electrodes[j-lendotprodcoeffs0].perCellLFP:
                        if not hasattr(electrodes[j], 'CellLFP'):
                            electrodes[j-lendotprodcoeffs0].CellLFP = []
                        electrodes[j-lendotprodcoeffs0].CellLFP.append(LFP)
                    electrodes[j-lendotprodcoeffs0].electrodecoeff = dotprodcoeffs[j]

    if to_file:
        el_LFP_file.close()


def _collect_geometry_neuron(cell):
    '''Loop over allseclist to determine area, diam, xyz-start- and
    endpoints, embed geometry to cell object'''


    areavec = np.zeros(cell.totnsegs)
    diamvec = np.zeros(cell.totnsegs)
    lengthvec = np.zeros(cell.totnsegs)

    xstartvec = np.zeros(cell.totnsegs)
    xendvec = np.zeros(cell.totnsegs)
    ystartvec = np.zeros(cell.totnsegs)
    yendvec = np.zeros(cell.totnsegs)
    zstartvec = np.zeros(cell.totnsegs)
    zendvec = np.zeros(cell.totnsegs)

    counter = 0

    #loop over all segments
    for sec in cell.allseclist:
        n3d = int(neuron.h.n3d())
        nseg = sec.nseg
        gsen2 = 1./2/nseg
        if n3d > 0:
            #create interpolation objects for the xyz pt3d info:
            L = np.zeros(n3d)
            x = np.zeros(n3d)
            y = np.zeros(n3d)
            z = np.zeros(n3d)
            for i in range(n3d):
                L[i] = neuron.h.arc3d(i)
                x[i] = neuron.h.x3d(i)
                y[i] = neuron.h.y3d(i)
                z[i] = neuron.h.z3d(i)

            #normalize as seg.x [0, 1]
            L /= sec.L

            #temporary store position of segment midpoints
            segx = np.zeros(nseg)
            for i, seg in enumerate(sec):
                segx[i] = seg.x

            #can't be >0 which may happen due to NEURON->Python float transfer:
            segx0 = (segx - gsen2).round(decimals=6)
            segx1 = (segx + gsen2).round(decimals=6)

            #fill vectors with interpolated coordinates of start and end points
            xstartvec[counter:counter+nseg] = np.interp(segx0, L, x)
            xendvec[counter:counter+nseg] = np.interp(segx1, L, x)

            ystartvec[counter:counter+nseg] = np.interp(segx0, L, y)
            yendvec[counter:counter+nseg] = np.interp(segx1, L, y)

            zstartvec[counter:counter+nseg] = np.interp(segx0, L, z)
            zendvec[counter:counter+nseg] = np.interp(segx1, L, z)

            #fill in values area, diam, length
            for seg in sec:
                areavec[counter] = neuron.h.area(seg.x)
                diamvec[counter] = seg.diam
                lengthvec[counter] = sec.L/nseg

                counter += 1

    #set cell attributes
    cell.xstart = xstartvec
    cell.ystart = ystartvec
    cell.zstart = zstartvec

    cell.xend = xendvec
    cell.yend = yendvec
    cell.zend = zendvec

    cell.area = areavec
    cell.diam = diamvec
    cell.length = lengthvec



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Main simulation script for setting up the model producing the data shown in
figure 4, 5 and 6 in:

    Multimodal modeling of neural network activity: computing LFP, ECoG, EEG and
    MEG signals with LFPy2.0
    Espen Hagen, Solveig Næss, Torbjørn V Ness, Gaute T Einevoll
    bioRxiv 281717; doi: https://doi.org/10.1101/281717

This example file is not suited for execution on laptops or desktop computers.
In the corresponding parameter file example_parallel_network_parameters.py there
is an option to set

    TESTING = True
    
which will set the number of neurons in each network population to one, which
facilitates testing for missing files or data. Otherwise, the full number of
neurons in each population will be created.

Various biophysically detailed neuron models required for these simulations
can be obtained from The Neocortical Microcircuit Collaboration Portal
(https://bbp.epfl.ch/nmc-portal/welcome) providing the cell models used in a
reconstruction of a rat somatosensory cortex column described in:

Markram H, Muller E, Ramaswamy S†, Reimann MW,  Abdellah M, Sanchez CA,
Ailamaki A, Alonso-Nanclares L, Antille N, Arsever S et al. (2015).
Reconstruction and Simulation of Neocortical Microcircuitry.
Cell 163:2, 456 - 492. doi: 10.1016/j.cell.2015.09.029

A tar file with all single-cell models zipped can be downloaded and unpacked by issuing:
$ wget https://bbp.epfl.ch/nmc-portal/documents/10184/7288948/hoc_combos_syn.1_0_10.allzips.tar
$ tar -xvf hoc_combos_syn.1_0_10.allzips.tar
$ cd hoc_combos_syn.1_0_10.allzips
$ unzip 'L4_PC_*.zip'
$ unzip 'L4_LBC_*.zip'
$ unzip 'L5_TTPC1_*.zip'
$ unzip 'L5_MC_*.zip'
$ cd -

Some preparatory steps has to be made in order to compile NMODL language
files used by the neuron models:

Set working dir
>>> import os
>>> import sys
>>> import neuron
>>> from glob import glob
>>> CWD = os.getcwd()
>>> NMODL = 'hoc_combos_syn.1_0_10.allmods'

Load some required neuron-interface files
>>> neuron.h.load_file("stdrun.hoc")
>>> neuron.h.load_file("import3d.hoc")

Load only some layer 5 pyramidal cell types
Define a list of the neuron models (defined in the parameter file)
>>> neurons = [os.path.join('hoc_combos_syn.1_0_10.allzips', 'L4_PC_cADpyr230_1'),
               os.path.join('hoc_combos_syn.1_0_10.allzips', 'L4_LBC_dNAC222_1'),
               os.path.join('hoc_combos_syn.1_0_10.allzips', 'L5_TTPC1_cADpyr232_1'),
               os.path.join('hoc_combos_syn.1_0_10.allzips', 'L5_MC_bAC217_1')]

Attempt to set up a folder with all unique mechanism mod files, compile, and
load them all. One synapse mechanism file is faulty and must be patched.
>>> if not os.path.isdir(NMODL):
>>>     os.mkdir(NMODL)
>>> for NRN in neurons:
>>>     for nmodl in glob(os.path.join(NRN, 'mechanisms', '*.mod')):
>>>         while not os.path.isfile(os.path.join(NMODL,
>>>                                               os.path.split(nmodl)[-1])):
>>>             if "win32" in sys.platform:
>>>                 os.system("copy {} {}".format(nmodl, NMODL))
>>>             else:
>>>                 os.system('cp {} {}'.format(nmodl,
>>>                                             os.path.join(NMODL,
>>>                                                      '.')))
>>> os.chdir(NMODL)
>>> diff = """319c319
<                 urand = scop_random(1)
---
>                 value = scop_random(1)
"""
>>> f = open('ProbGABAAB_EMS.patch', 'w')
>>> f.writelines(diff)
>>> f.close()
>>> os.system('patch ProbGABAAB_EMS.mod ProbGABAAB_EMS.patch')
>>> if "win32" in sys.platform:
>>>     warn("no autompile of NMODL (.mod) files on Windows. " 
>>>          + "Run mknrndll from NEURON bash in the folder %s and rerun example script" % NMODL)
>>> else:
>>>     os.system('nrnivmodl')        
>>> os.chdir(CWD)


An example job script set up using the SLURM workload management software
(https://slurm.schedmd.com/) on a compute cluster is provided in the file
example_parallel_network.job. This job script asks for exclusive access to
24 nodes with 24 physical CPU cores each. Adjust accordingly for other clusters.
The job can be submitted issuing:

    $ sbatch example_parallel_network.job


Execution example:

    $ mpirun -np 1152 python example_parallel_network.py

Output is stored in the folder ./example_parallel_network_output, which is set
in the parameter file. Adjust accordingly to use the work area on your cluster. 
Some pdf figures are saved as example_parallel_network*.pdf,
showing somatic responses of different cells on some MPI processes

The script uses some plotting routines from the file
example_parallel_network_plotting.py


If the simulation finished successfully, the figures 4-6 from the manuscript
can be generated by issuing:

    $ python figure_4.py
    $ python figure_5.py
    $ python figure_6.py


Copyright (C) 2018 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
'''
from __future__ import division
from mpi4py import MPI
import neuron
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.collections import PolyCollection
from matplotlib.gridspec import GridSpec
import numpy as np
from distutils.version import LooseVersion
import h5py
if LooseVersion(h5py.version.hdf5_version) < LooseVersion('1.8.16'):
    raise ImportError('h5py uses HDF5 v{}: v1.8.16 or newer required'.format(
        h5py.version.hdf5_version))
import os
from time import time
import LFPy
from example_parallel_network_plotting import decimate


# set up MPI environment
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

# set default plotting parameters
fontsize=8
plt.rcParams.update({
    'axes.xmargin': 0.0,
    'axes.ymargin': 0.0,
    'axes.labelsize': fontsize,
    'axes.titlesize': fontsize,
    'axes.titleweight': fontsize,
    'figure.titlesize': fontsize,
    'font.size': fontsize,
    'legend.fontsize': fontsize,
})


# tic toc
tic = time()

# avoid same sequence of random numbers from numpy and neuron on each RANK,
# e.g., in order to draw unique cell locations and random synapse activation
# times
GLOBALSEED = 1234
np.random.seed(GLOBALSEED + RANK)


################################################################################
# Main simulation procedure
################################################################################

if __name__ == '__main__':
    # Remove cells from previous script executions
    neuron.h('forall delete_section()')
    
    ############################################################################
    # Simulation control and parameters
    ############################################################################
    # import LFPy.NetworkCell and Network classes
    from LFPy import NetworkCell, Network

    # tic-toc
    if RANK == 0:
        initialization_time = time() - tic
        print('Initialization in {} seconds'.format(initialization_time))
    tic = time()

    # import main parameters dictionary for simulation
    from example_parallel_network_parameters import PSET

    # set up file destination
    if not os.path.isdir(PSET.OUTPUTPATH):
        if RANK == 0:
            os.mkdir(PSET.OUTPUTPATH)
    COMM.Barrier()

    if RANK == 0:
        parameters_time = time() - tic
        print('Parameters in {} seconds'.format(parameters_time))
    tic = time()


    ############################################################################
    # Create population, provide noisy input to each cell, connect network
    ############################################################################
    # create network object instance
    network = Network(dt=PSET.dt, tstop=PSET.tstop, v_init=PSET.v_init,
                      celsius=PSET.celsius,
                      OUTPUTPATH=PSET.OUTPUTPATH)
    # create populations iteratively
    for name, pop_args, rotation_args, POP_SIZE in zip(
        PSET.populationParameters['me_type'],
        PSET.populationParameters['pop_args'],
        PSET.populationParameters['rotation_args'],
        (PSET.populationParameters['POP_SIZE']*PSET.POPSCALING).astype(int)):
        network.create_population(CWD=PSET.CWD, CELLPATH=PSET.CELLPATH,
                                  Cell=NetworkCell, POP_SIZE=POP_SIZE,
                                  name=name, cell_args=PSET.cellParameters[name],
                                  pop_args=pop_args, rotation_args=rotation_args)

    # tic-toc
    if RANK == 0:
        create_population_time = time() - tic
        print('Populations initialized in {} seconds'.format(create_population_time))
    tic = time()

    
    # Sync MPI threads as populations may take a different amount of
    # time across RANKs. All neurons must have been created before connections
    # are made
    COMM.Barrier()

    # 
    # # Attach current stimulus to the soma of the cell with gid 0.
    # if False:
    #     for name in PSET.populationParameters['me_type']:
    #         for cell in network.populations[name].cells:
    #             if cell.gid == 0:
    #                 LFPy.StimIntElectrode(cell,
    #                                       amp = 0.4,
    #                                       **PSET.PointProcParams)
    
    # create for each cell in each population some external input with Poisson
    # statistics using NEURON's NetStim device (controlled using LFPy.Synapse)
    for m_type, me_type, section, rho, f, synparams, weightfun, weightargs in zip(
                        PSET.populationParameters['m_type'],
                        PSET.populationParameters['me_type'],
                        PSET.populationParameters['extrinsic_input_section'],
                        PSET.populationParameters['extrinsic_input_density'],
                        PSET.populationParameters['extrinsic_input_frequency'],
                        PSET.connParamsExtrinsic['synparams'],
                        PSET.connParamsExtrinsic['weightfuns'],
                        PSET.connParamsExtrinsic['weightargs'],
                        ):
        for cell in network.populations[me_type].cells:
            idx = cell.get_rand_idx_area_norm(section=section,
                                              nidx=int(cell.area[cell.get_idx(section)].sum() * rho))
            for i in idx:
                syn = LFPy.Synapse(cell=cell, idx=i,
                                   syntype=PSET.connParamsExtrinsic['syntype'],
                                   weight=weightfun(**weightargs),
                                   **synparams)
                syn.set_spike_times_w_netstim(interval=1000. / f)
        
    
    
    # connect pre and post-synaptic populations with some connectivity and
    # weight of connections and other connection parameters:
    total_conncount = 0
    total_syncount = 0
    for i, pre in enumerate(PSET.populationParameters['me_type']):
        for j, post in enumerate(PSET.populationParameters['me_type']):
            # boolean connectivity matrix between pre- and post-synaptic neurons
            # in each population (postsynaptic on this RANK)
            connectivity = network.get_connectivity_rand(pre=pre, post=post,
                                    connprob=PSET.connParams['connprob'][i][j])
            
            # connect network
            (conncount, syncount) = network.connect(
                            pre=pre, post=post,
                            connectivity=connectivity,
                            syntype=PSET.connParams['syntypes'][i][j],
                            synparams=PSET.connParams['synparams'][i][j],
                            weightfun=PSET.connParams['weightfuns'][i][j],
                            weightargs=PSET.connParams['weightargs'][i][j],
                            delayfun=PSET.connParams['delayfuns'][i][j],
                            delayargs=PSET.connParams['delayargs'][i][j],
                            multapsefun=PSET.connParams['multapsefuns'][i][j],
                            multapseargs=PSET.connParams['multapseargs'][i][j],
                            syn_pos_args=PSET.connParams['syn_pos_args'][i][j],
                            save_connections=PSET.save_connections,
                            )
            total_conncount += conncount
            total_syncount += syncount
    
    # tic-toc
    if RANK == 0:
        create_connections_time = time() - tic
        print('Network build finished with {} connections and {} synapses in {} seconds'.format(
            total_conncount, total_syncount, create_connections_time))
    tic = time()


    ############################################################################
    # Set up extracellular electrode
    ############################################################################
    if PSET.COMPUTE_LFP:
        electrode = LFPy.RecExtElectrode(**PSET.electrodeParams)
    else:
        electrode = None
    
    if PSET.COMPUTE_ECOG:
        ecog_electrode = LFPy.RecMEAElectrode(**PSET.ecogParameters)
        electrode = [electrode, ecog_electrode]


    ############################################################################
    # Recording of additional variables
    ############################################################################
    if RANK == 0:
        network.t = neuron.h.Vector()
        network.t.record(neuron.h._ref_t)
    else:
        network.t = None

    
    ############################################################################
    # run simulation, gather results across all RANKs
    ############################################################################
    # Assert that connect routines has finished across RANKS before starting
    # main simulation procedure
    COMM.Barrier()
    if RANK == 0:
        print('running simulation....')
    SPIKES, SUMMED_OUTPUT, CURRENT_DIPOLE_MOMENT = network.simulate(
                              electrode=electrode,
                              rec_current_dipole_moment=PSET.COMPUTE_P,
                              rec_pop_contributions=PSET.rec_pop_contributions,
                              **PSET.NetworkSimulateArgs)
    # note: OUTPUT is a structured array if
    #   PSET.COMPUTE_LFP = True and PSET.COMPUTE_P is False and
    #   PSET.rec_pop_contributions is False.
    #   If PSET.COMPUTE_P == True and/or PSET.rec_pop_contributions == True,
    #   then output is a tuple (OUTPUT, current_dipole_moment)
    if RANK == 0:
        run_simulation_time = time() - tic
        print('Simulations finished in {} seconds'.format(run_simulation_time))
    tic = time()

    

    ############################################################################
    # save simulated output to file to allow for offline plotting
    ############################################################################

    if RANK == 0:
        f = h5py.File(os.path.join(PSET.OUTPUTPATH,
                                   'example_parallel_network_output.h5'),
                      'w')
        if PSET.COMPUTE_P:
            f['CURRENT_DIPOLE_MOMENT'] = CURRENT_DIPOLE_MOMENT
        if PSET.COMPUTE_LFP:
            # save all extracellular traces
            f['SUMMED_OUTPUT'] = SUMMED_OUTPUT[0]
        if PSET.COMPUTE_ECOG:
            # save extracellular potential on top of cortex
            f['SUMMED_ECOG'] = SUMMED_OUTPUT[1]
        # all spikes
        grp = f.create_group('SPIKES')
        dtype = h5py.special_dtype(vlen=np.dtype('float')) # variable length datatype for spike time arrays
        for i, name in enumerate(PSET.populationParameters['me_type']):
            subgrp = grp.create_group(name)
            if len(SPIKES['gids'][i]) > 0:
                subgrp['gids'] = np.array(SPIKES['gids'][i]).flatten()
                dset = subgrp.create_dataset('times', (len(SPIKES['gids'][i]),),
                                             dtype=dtype)
                for j, spt in enumerate(SPIKES['times'][i]):
                    dset[j] = spt
            else:
                subgrp['gids'] = []
                subgrp['times'] = []
        f.close()
    
    COMM.Barrier()

    # clean up namespace
    if PSET.COMPUTE_LFP:
        del SUMMED_OUTPUT
    # del neurons, spikes

    # tic toc
    if RANK == 0:
        saving_data_time = time() - tic
        print('Wrote output files in {} seconds'.format(saving_data_time))
    tic = time()


    # create logfile recording time in seconds for different simulation steps
    # (initialization, parameters, simulation etc.)
    if RANK == 0:
        logfile = open(os.path.join(PSET.OUTPUTPATH, 'log.txt'), 'w')
        logfile.write('initialization {}\n'.format(initialization_time))
        logfile.write('parameters {}\n'.format(parameters_time))
        logfile.write('population {}\n'.format(create_population_time))
        logfile.write('connections {}\n'.format(create_connections_time))
        logfile.write('simulation {}\n'.format(run_simulation_time))
        logfile.write('save {}\n'.format(saving_data_time))
        logfile.close()


if __name__ == '__main__':    
    ############################################################################
    # Plot simulated output (relies on Network class instance)
    ############################################################################
    
    T = (PSET.TRANSIENT, PSET.tstop)
    
    colors = [plt.get_cmap('Set1', PSET.populationParameters.size)(i)
              for i in range(PSET.populationParameters.size)]
    
     # don't want thousands of figure files:
    PLOTRANKS = np.arange(0, SIZE, 16) if SIZE >= 48 else np.arange(SIZE)

    if RANK in PLOTRANKS:
        fig = plt.figure(figsize=(20, 10)) 
        fig.subplots_adjust(left=0.2)
        
        nrows = np.sum([len(population.gids)
                        for population in network.populations.values()])
        ncols = 1
            
        gs = GridSpec(nrows=nrows, ncols=ncols)
        
        fig.suptitle('RANK {}'.format(RANK))
        counter = 0
        # somatic traces
        tvec = np.arange(PSET.tstop / PSET.dt + 1)*PSET.dt
        tinds = tvec >= PSET.TRANSIENT
        for i, name in enumerate(PSET.populationParameters['me_type']):
            population = network.populations[name]
            for j, cell in enumerate(population.cells):
                ax = fig.add_subplot(gs[counter, 0])
                if counter == 0:
                    ax.set_title('somatic voltages')
                ax.plot(tvec[tinds][::PSET.decimate_q],
                        decimate(cell.somav[tinds],
                                 q=PSET.decimate_q),
                        color=colors[i], lw=1.5, label=name)
                ax.set_ylabel('gid {}'.format(population.gids[j]),
                                         rotation='horizontal', labelpad=30)
                ax.axis(ax.axis('tight'))
                ax.set_ylim(-90, -20)
                ax.legend(loc='best')
                ax.yaxis.set_major_locator(MaxNLocator(nbins=3))
                counter += 1
                if counter == nrows:
                    ax.set_xlabel('time (ms)')
                else: 
                    ax.set_xticklabels([])
        
        # save figure output
        fig.savefig(os.path.join(PSET.OUTPUTPATH,
                                 'example_parallel_network_RANK_{}.pdf'.format(
                                    RANK)),
                    bbox_inches='tight')
        plt.close(fig)
    

    # make an illustration of the different populations on each RANK.
    if RANK in PLOTRANKS: # don't want thousands of figure files
        
        # figure out which populations has at least one cell on this RANK
        local_me_types = []
        for i, name in enumerate(PSET.populationParameters['me_type']):
            if len(network.populations[name].gids) >= 1:
                local_me_types += [name]
        
        fig, axes = plt.subplots(1, len(local_me_types)+1,
                                 figsize=((len(local_me_types)+1)*5, 10),
                                 sharey=True, sharex=True)    
        ax = axes[0]
        # plot electrode contact points
        ax.plot(PSET.electrodeParams['x'], PSET.electrodeParams['z'],
                'ko', markersize=5)
        
        # plot cell geometries
        for i, name in enumerate(PSET.populationParameters['me_type']):
            population = network.populations[name]
            zips = []
            for cell in population.cells:
                for x, z in cell.get_idx_polygons(projection=('x', 'z')):
                    zips.append(list(zip(x, z)))            
            polycol = PolyCollection(zips,
                                     edgecolors=colors[i],
                                     linewidths=0.01,
                                     facecolors=colors[i],
                                     label=name,
                                     )
            ax.add_collection(polycol)
        ax.set_xlim(-400, 400)
        axis = ax.axis()
        ax.hlines(np.r_[0., -PSET.layer_data['thickness'].cumsum()],
                  axis[0], axis[1], 'k', lw=0.5)
        ax.set_xticks([-400, 0, 400])
        ax.set_xlabel(r'x ($\mu$m)')    
        ax.set_ylabel(r'z ($\mu$m)')    
        ax.set_title('network populations')        
        
        # for i, (name, population) in enumerate(network.populations.items()):
        j = 1 #counter
        for i, name in enumerate(PSET.populationParameters['me_type']):
            if name in local_me_types:
                population = network.populations[name]
                ax = axes[j]
                # plot electrode contact points
                ax.plot(PSET.electrodeParams['x'], PSET.electrodeParams['z'],
                        'ko', markersize=5)
                
                # plot cell geometries and synapse locations
                zips = []
                synpos_e = []
                synpos_i = []
                for cell in population.cells:
                    for x, z in cell.get_idx_polygons(projection=('x', 'z')):
                        zips.append(list(zip(x, z)))
                    for idx, syn in zip(cell.synidx, cell.netconsynapses):
                        if hasattr(syn, 'e') and syn.e > -50:
                            synpos_e += [[cell.xmid[idx], cell.zmid[idx]]]
                        else:
                            synpos_i += [[cell.xmid[idx], cell.zmid[idx]]]
    
                polycol = PolyCollection(zips,
                                         edgecolors=colors[i],
                                         linewidths=0.01,
                                         facecolors=colors[i],
                                         label=name,
                                         )
                ax.add_collection(polycol)
                synpos_e = np.array(synpos_e)
                synpos_i = np.array(synpos_i)
                if synpos_e.size > 0:
                    ax.plot(synpos_e[:, 0], synpos_e[:, 1], 'r.',
                            markersize=3, zorder=1)
                if synpos_i.size > 0:
                    ax.plot(synpos_i[:, 0], synpos_i[:, 1], 'b.',
                            markersize=1.5, zorder=2)
                ax.set_xlim(-400, 400)
                axis = ax.axis()
                ax.hlines(np.r_[0., -PSET.layer_data['thickness'].cumsum()],
                          axis[0], axis[1], 'k', lw=0.5)
                ax.set_xticks([-400, 0, 400])
                ax.set_xlabel(r'x ($\mu$m)')
                ax.set_title(name)
                j += 1 #counter 
            
        # save figure output
        fig.savefig(os.path.join(PSET.OUTPUTPATH,
                    'example_parallel_network_populations_RANK_{}.pdf'.format(
                        RANK)),
                    bbox_inches='tight')
        plt.close(fig)  

    
    ############################################################################
    # customary cleanup of object references - the psection() function may not
    # write correct information if NEURON still has object references in memory,
    # even if Python references has been deleted. It will also allow the script
    # to be run in successive fashion.
    ############################################################################
    network.pc.gid_clear() # allows assigning new gids to threads 
    for population in network.populations.values():
        for cell in population.cells:
            cell = None
        population.cells = None
        population = None
    network = None
    neuron.h('forall delete_section()')

